#include <stdio.h>
#include <math.h>
#include "cuda.h"
#include "../common/CudaHelper.h"
#include "mex.h"

#define BLOCK_SIZE 128
#define MAX_CHANNELS 192
#define MAX_BEAMS 40

texture<float,2,cudaReadModeNormalizedFloat> tex;
texture<float,2,cudaReadModeNormalizedFloat> texApo;


__constant__ float C1_gpu[MAX_CHANNELS];
__constant__ float C3_gpu[MAX_CHANNELS];
__constant__ float C4_gpu[MAX_BEAMS];

__global__ void delay_and_sum_kernel( float *out_data, int channels, int samples )
{
	int idx = blockIdx.x*BLOCK_SIZE+threadIdx.x;
	if(idx < samples){
		float i = idx;
		float out = 0.0f;
		int beam = blockIdx.y;
		float y = 0.5f + channels*beam;
		float apoY = 0.5f;
		float y_max = channels*(beam+1);
		int n = 0;
		do{
			out += tex2D(texApo, 0.5f + i,apoY) *
				   tex2D(tex, 0.5f + i/2 + sqrt(i*i/4 + C1_gpu[n] - i*C3_gpu[n]*C4_gpu[beam]),y);
			y += 1.0f;
			apoY += 1.0f;
			n += 1;
		} while( y < y_max );
		idx += beam*samples;
		out_data[idx] = out;	
	}
}

void checkLaunch(){
	cudaThreadSynchronize();                      
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) {                                          
        fprintf(stderr, "failed : %s.\n",
               cudaGetErrorString( err) );
    } else
		printf("success\n");
}

void delay_and_sum(short *in_data, float *out_data, short *apo, double *a_c, double f_s, double c_0, 
				   double *theta, unsigned int samples, unsigned int channels, int beams)
{

	// allocate memory
	float *out_data_gpu;
	cudaMalloc( (void **) &out_data_gpu,sizeof(float)*samples*beams);


	float C1[MAX_CHANNELS] = {0};
	float C3[MAX_CHANNELS] = {0};
	float C4[MAX_BEAMS] = {0};

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(16, 0, 0, 0, cudaChannelFormatKindSigned);

	// allocate device memory in_data for texture
	cudaArray* in_data_gpu;
	CUDA_SAFE_CALL(cudaMallocArray( &in_data_gpu, &channelDesc, samples, channels*beams )); 

	// copy host data to device array
	CUDA_SAFE_CALL(cudaMemcpyToArray( in_data_gpu, 0, 0, in_data, sizeof(short)*samples*channels*beams, cudaMemcpyHostToDevice));

	// set texture parameters
	tex.addressMode[0] = cudaAddressModeClamp;
	tex.addressMode[1] = cudaAddressModeClamp;
	tex.filterMode = cudaFilterModeLinear;
	tex.normalized = false;

	// Bind the array to the texture
	CUDA_SAFE_CALL(cudaBindTextureToArray(tex, in_data_gpu, channelDesc));


	// init APODIZE texture
	cudaChannelFormatDesc channelDescApo = cudaCreateChannelDesc(16, 0, 0, 0, cudaChannelFormatKindSigned);
	cudaArray* apo_gpu;
	// allocate device for apodize texture
	cudaMallocArray( &apo_gpu, &channelDescApo, samples, channels ); 

	// set texture parameters
	texApo.addressMode[0] = cudaAddressModeClamp;
	texApo.addressMode[1] = cudaAddressModeClamp;
	texApo.filterMode = cudaFilterModePoint;
	texApo.normalized = false;

	// copy host data to device array
	CUDA_SAFE_CALL(cudaMemcpyToArray( apo_gpu, 0, 0, apo, sizeof(short)*samples*channels, cudaMemcpyHostToDevice));

	// Bind the array to the texture
	cudaBindTextureToArray(texApo, apo_gpu, channelDescApo);




	// calculate C1, C3 and C4
	double lambda = f_s/c_0;
	for (int j = 0; j < channels; j++) {
		double temp = a_c[j]*lambda;
		C1[j] = (float)( temp*temp );
		C3[j] = (float)temp;
	}	
	for (int b = 0; b < beams; b++) {
		C4[b] = (float)sin(theta[b]);
	}

	// copy C1, C3 and C4 to device
	cudaMemcpyToSymbol( C1_gpu, C1, sizeof(float)*MAX_CHANNELS);
	cudaMemcpyToSymbol( C3_gpu, C3, sizeof(float)*MAX_CHANNELS);
	cudaMemcpyToSymbol( C4_gpu, C4, sizeof(float)*MAX_BEAMS);



	// run kernel
	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( (samples+BLOCK_SIZE-1)/BLOCK_SIZE,beams,1 );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);

	TIC();

	int reps = 100;
	for(int n=0;n<reps;n++)
		delay_and_sum_kernel<<<dimGrid,dimBlock>>>( out_data_gpu, channels, samples );

	float wallTime;
	TOC(wallTime);

	checkLaunch();

	float dataTransfered = (float)reps*beams*samples*(channels*sizeof(short) + sizeof(float));
	printf("time avg per beam: %.2f(ms)\n",1000*wallTime/(reps*beams));
	printf("time available per beam: %.2f(ms)\n",1000*samples/f_s);
	printf("Speed: %.2fx\n",(beams*samples/f_s)/(wallTime/reps));
	printf("Datatransfer: approx. %.2fGB/s\n",dataTransfered/(wallTime*1024*1024*1024));

	// copy result to host memory
	CUDA_SAFE_CALL(cudaMemcpy( out_data, out_data_gpu, sizeof(float)*samples*beams, cudaMemcpyDeviceToHost));
	
	// clean up
	CUDA_SAFE_CALL(cudaUnbindTexture(tex));
	CUDA_SAFE_CALL(cudaUnbindTexture(texApo));
	CUDA_SAFE_CALL(cudaFreeArray(in_data_gpu));
	CUDA_SAFE_CALL(cudaFreeArray(apo_gpu));

	CUDA_SAFE_CALL(cudaFree(out_data_gpu));
}


int length(const mxArray *vector){
	if( mxGetM(vector) == 1 ){
		return(mxGetN(vector));
	} else if( mxGetN(vector) == 1 ) {
		return(mxGetM(vector));
	} else {
		mexErrMsgTxt("Not a vector");
		return 0;
	}
}

/* Gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// Get parameters
	//
	// function [out_data] = TEXF_APO( in_data, a_c, f_s, c_0, theta, apo )

	const mxArray *mx_in_data =	prhs[0], 
		          *mx_a_c =		prhs[1], 
				  *mx_f_s =		prhs[2], 
				  *mx_c_0 =		prhs[3],
				  *mx_theta =	prhs[4],
				  *mx_apo =		prhs[5];

	// check validity of parameters
	if (nrhs != 6)
		mexErrMsgTxt("Wrong input parameters. TEXF( in_data, a_c, f_s, c_0, theta, apo )");

	int channels = length(mx_a_c);
	int beams = length(mx_theta);
	int samples = mxGetM(mx_in_data);
	
	if( mxGetN(mx_in_data) != channels*beams )
		mexErrMsgTxt("Number of columns of in_data must equal length of a_c times length of theta");

	if( beams > MAX_BEAMS )
		mexErrMsgTxt("Number of beams must be no more than 40");

	if( channels > MAX_CHANNELS )
		mexErrMsgTxt("Number of channels must be no more than 192");

	if( mxGetM(mx_apo) != samples )
		mexErrMsgTxt("Number of rows in apo must be same as in_data");

	if( mxGetN(mx_apo) != channels )
		mexErrMsgTxt("Number of columns in apo must be same as number of channels");

	printf("%d samples x %d channels x %d beams\n",samples,channels,beams);

	// get scalars
	double f_s = mxGetScalar(mx_f_s);
	double c_0 = mxGetScalar(mx_c_0);

	// get pointers
	double *in_data = mxGetPr(mx_in_data);
	double *apo = mxGetPr(mx_apo);
	double *a_c = mxGetPr(mx_a_c);
	double *theta = mxGetPr(mx_theta);




	// allocate memory
	short *in_datas = (short*)mxMalloc(sizeof(short)*samples*channels*beams);
	short *apos = (short*)mxMalloc(sizeof(short)*samples*channels);
	float *out_dataf = (float*)mxMalloc(sizeof(float)*samples*beams);

	// convert in_data. double -> short
	for( int i = 0; i < samples*channels*beams; i++)
		in_datas[i] = (short)(in_data[i]*0x7fff);

	// convert apo. double -> short
	for( int i = 0; i < samples*channels; i++)
		apos[i] = (short)(apo[i]*0x7fff);

	// do it
	delay_and_sum(in_datas, out_dataf, apos, a_c, f_s, c_0, theta, samples, channels, beams);

	// create result matrix
	plhs[0] = mxCreateDoubleMatrix(samples, beams, mxREAL);
	double *out_data = mxGetPr(plhs[0]);

	// convert out_data. float -> double
	for (int i = 0; i < samples*beams; i++)
			out_data[i] = (double)out_dataf[i];

	// free memory
	mxFree(out_dataf);
	mxFree(apos);
	mxFree(in_datas);

}