#include <stdio.h>
#include <string.h>
#include "cuda.h"
#include "mex.h"
#include "../common/CudaHelper.h"

#define BLOCK_SIZE 256
// DATA_PER_BLOCK: each thread handles two samples
#define DATA_PER_BLOCK (2*BLOCK_SIZE)

// TAPS must be odd number
#define TAPS 11

__constant__ float c_coeffs[3][TAPS];

texture<float2, 1, cudaReadModeNormalizedFloat> tex_short2;

__global__ void upsample4x_kernel( float4 *out_data, short channel_pitch_in, short channel_pitch_out )
{
	extern __shared__ float shared[];

	const unsigned int ti = threadIdx.x;
	const unsigned int i_out = blockIdx.y*channel_pitch_out + blockIdx.x*BLOCK_SIZE*2 + ti;
	const unsigned int i_in = blockIdx.y*channel_pitch_in + blockIdx.x*BLOCK_SIZE + ti;
	float4 out;

	// load data into shared memory
	float2 in = tex1Dfetch(tex_short2, i_in);
	shared[ti*2]   = in.x; 
	shared[ti*2+1] = in.y;

	if(ti < TAPS-1){
		in = tex1Dfetch(tex_short2, BLOCK_SIZE + i_in);
		shared[BLOCK_SIZE*2 + ti*2]   = in.x; 
		shared[BLOCK_SIZE*2 + ti*2+1] = in.y;
	}

	__syncthreads();


	out.x = shared[ti + TAPS/2];	// D = 0
	out.y = 0.0f; out.z = 0.0f; out.w = 0.0f;
	for( int n = 0; n<TAPS; n++)
		out.y += c_coeffs[0][n]*shared[ti+TAPS-1-n]; // D = -0.25
	for( int n = 0; n<TAPS; n++)
		out.z += c_coeffs[1][n]*shared[ti+TAPS-1-n]; // D = -0.5
	for( int n = 0; n<TAPS; n++)
		out.w += c_coeffs[2][n]*shared[ti+TAPS-1-n]; // D = -0.75

	out_data[i_out] = out;

	out.x = shared[BLOCK_SIZE + ti + TAPS/2];	// D = 0
	out.y = 0.0f; out.z = 0.0f; out.w = 0.0f;
	for( int n = 0; n<TAPS; n++)
		out.y += c_coeffs[0][n]*shared[BLOCK_SIZE + ti+TAPS-1-n]; // D = -0.25
	for( int n = 0; n<TAPS; n++)
		out.z += c_coeffs[1][n]*shared[BLOCK_SIZE + ti+TAPS-1-n]; // D = -0.5
	for( int n = 0; n<TAPS; n++)
		out.w += c_coeffs[2][n]*shared[BLOCK_SIZE + ti+TAPS-1-n]; // D = -0.75
	
	out_data[BLOCK_SIZE + i_out] = out;
}


void upsample4x(short *in_data, float *out_data, float *coeffs, unsigned int samples,
				unsigned int filter_len, unsigned int padding, int channels)
{
	// allocate memory
	float4 *out_data_gpu;
	short *in_data_gpu;
	int channel_pitch_in = padding + samples + filter_len - 1;
	int in_data_len = sizeof(short)*channel_pitch_in*channels;
	int channel_pitch_out = padding + samples;
	int out_data_len = sizeof(float4)*channel_pitch_out*channels;
	cudaMalloc( (void **) &in_data_gpu, in_data_len);
	cudaMalloc( (void **) &out_data_gpu, out_data_len);

	// copy to device
	cudaMemcpyToSymbol( c_coeffs, coeffs, sizeof(float)*filter_len*3);
	cudaMemcpy( in_data_gpu, in_data, in_data_len, cudaMemcpyHostToDevice);

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(16, 16, 0, 0, cudaChannelFormatKindSigned);
	CUDA_SAFE_CALL( cudaBindTexture(0, tex_short2, in_data_gpu, channelDesc, in_data_len ));

	int sharedMemSize = sizeof(float)*(DATA_PER_BLOCK+filter_len-1);

	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( (samples+padding)/DATA_PER_BLOCK,channels,1 );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);

    TIC();

	int reps = 100;
	for(int n=0;n<reps;n++){
		upsample4x_kernel<<<dimGrid,dimBlock,sharedMemSize>>>( out_data_gpu, channel_pitch_in/2, channel_pitch_out );
	}


	float time_taken;
	TOC(time_taken);

	printf("avg. time: %.3f(ms)\n",1000*time_taken/reps);

	double datatransferred = (double)reps*( dimGrid.x*dimGrid.y*sizeof(short2)*dimBlock.x + 
		out_data_len );
	printf("Transferrate: (at least) %.2fGiB/s\n",datatransferred/(time_taken*1024*1024*1024));

	double total_samples = (double)reps*samples;
	printf("Filterrate: %.2f*10^6 samples per channel per second\n",total_samples/(time_taken*1000*1000));

	double FLOPS = (double)reps*samples*filter_len*2*3*channels;
	printf("Floating point operation rate: %.2fGFLOPS\n",FLOPS/(time_taken*1000*1000*1000));


	// copy result to host memory
	cudaMemcpy( out_data, out_data_gpu, out_data_len, cudaMemcpyDeviceToHost);
	

	// clean up

	cudaFree(out_data_gpu);
	cudaFree(in_data_gpu);
}


/* Gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// Get parameters
	//
	// function [out_data] = Upsample4x( in_data, coeffs )

	const mxArray *mx_in_data =	prhs[0], 
		          *mx_coeffs =	prhs[1];

	// check validity of parameters
	if (nrhs != 2)
		mexErrMsgTxt("Wrong input parameters. Upsample4x( in_data, coeffs )\n");

	if( mxGetN(mx_coeffs) != 3 || mxGetM(mx_coeffs) != TAPS )
		mexErrMsgTxt("Wrong coeffs size. Must be TAPSx3\n");

	int samples = mxGetM(mx_in_data);
	int channels = mxGetN(mx_in_data);

	printf("%d channels x %d samples\n",channels,samples);

	double *in_data = mxGetPr(mx_in_data);
	double *coeffs = mxGetPr(mx_coeffs);

	unsigned int filter_len = TAPS;

	// calculate padding
	int padding = ((samples + DATA_PER_BLOCK - 1)/(DATA_PER_BLOCK))*DATA_PER_BLOCK - samples;
	int channel_pitch = (padding + samples + filter_len - 1);
	printf("filter length = %d   data length = %d   padding = %d \n",filter_len, samples, padding);

	// allocate memory
	short *in_datas = (short*)mxMalloc(sizeof(short)*channel_pitch*channels);
	float *out_dataf = (float*)mxMalloc(sizeof(float)*4*(padding + samples)*channels);
	float *coeffsf = (float*)mxMalloc(sizeof(float)*filter_len*3);
	
	for( int j = 0; j < channels; j++){
		// pad with zeros at beginning
		for( int i = 0; i < (filter_len/2); i++)
			in_datas[j*channel_pitch + i] = 0;

		// convert in_data. double -> float
		for( int i = 0; i < samples; i++)
			in_datas[j*channel_pitch + i + filter_len/2] = (short)(0x7FFF*in_data[j*samples + i]);
		
		// pad with zeros at end
		for( int i = 0; i < padding+filter_len/2; i++)
			in_datas[j*channel_pitch + i + filter_len/2 + samples] = 0;
	}

	// convert coeffs. double -> float
	for( int j = 0; j < 3; j++)
		for( int i = 0; i < filter_len; i++)
			coeffsf[j*filter_len + i] = (float)coeffs[j*filter_len + i];	

	// do it
	upsample4x(in_datas, out_dataf, coeffsf, samples, filter_len, padding, channels);

	// create result matrix
	plhs[0] = mxCreateDoubleMatrix(samples*4, channels, mxREAL);
	double *out_data = mxGetPr(plhs[0]);

	// convert out_data. float -> double
	for (int j = 0; j < channels; j++)
		for (int i = 0; i < samples*4; i++)
			out_data[j*samples*4 + i] = (double)out_dataf[j*(samples+padding)*4+i];

	// free memory
	mxFree(coeffsf);
	mxFree(out_dataf);
	mxFree(in_datas);
}
