#include <stdio.h>
#include <math.h>
#include "cuda.h"
#include "mex.h"
#include "../common/CudaHelper.h"


#define MAX_THREADS 512
#define MAX_BLOCKS 65536


__global__ void
bandwidthTestKernel( float* g_idata, clock_t* timer) 
{
	extern __shared__ float shared[];
	const unsigned int i = (threadIdx.x + threadIdx.y*blockDim.x
										+ threadIdx.z*blockDim.y*blockDim.x
										+ blockIdx.x *blockDim.y*blockDim.x*blockDim.z);
//	clock_t start_t = clock();
	shared[threadIdx.x] = g_idata[i];
	timer[i] = clock();
}


int length(const mxArray *vector){
	if( mxGetM(vector) == 1 ){
		return(mxGetN(vector));
	} else if( mxGetN(vector) == 1 ) {
		return(mxGetM(vector));
	} else {
		mexErrMsgTxt("Not a vector");
		return 0;
	}
}

void checkLaunch(){
	cudaThreadSynchronize();                      
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) {                                          
        fprintf(stderr, "failed : %s.\n",
               cudaGetErrorString( err) );
    } else
		printf("success\n");
}

/* Gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// Get parameters
	//
	// function [transferRate] = clockTest( threadsx, threadsy, threadsz, blocks, reps )

	const mxArray *mx_threadsx = prhs[0], 
				  *mx_threadsy = prhs[1], 
				  *mx_threadsz = prhs[2], 
		          *mx_blocks =	 prhs[3],
		          *mx_reps =	 prhs[4];

	// check validity of parameters
	if (nrhs != 5)
		mexErrMsgTxt("Wrong input parameters. clockTest( threadsx, threadsy, threadsz, blocks, reps )");

	int threadsx = (int)mxGetScalar(mx_threadsx);
	int threadsy = (int)mxGetScalar(mx_threadsy);
	int threadsz = (int)mxGetScalar(mx_threadsz);
	int threads = threadsx*threadsy*threadsz;
	int blocks = (int)mxGetScalar(mx_blocks);
	int reps = (int)mxGetScalar(mx_reps);

	plhs[0] = mxCreateDoubleMatrix(1, threads * blocks*reps, mxREAL);
	double *timer = mxGetPr(plhs[0]);

	// allocate memory
	clock_t *g_odata;
	float *g_idata;
	CUDA_SAFE_CALL(cudaMalloc( (void **) &g_idata,sizeof(float)*MAX_THREADS*MAX_BLOCKS));

	int odataSize = sizeof(clock_t)*threads*blocks;
	CUDA_SAFE_CALL(cudaMalloc( (void **) &g_odata, odataSize*reps));

	dim3  dimBlock( threadsx, threadsy, threadsz );
	dim3  dimGrid( blocks );

	TIC();

	for(int i = 0; i < reps; i++ )
		bandwidthTestKernel<<<dimGrid,dimBlock, (int)threads*4>>>(g_idata, &g_odata[i*threads * blocks]);
	
	float wallTime;
	TOC( wallTime );
	checkLaunch();
	// copy result to host memory
	clock_t *odata = (clock_t *)mxMalloc(odataSize*reps);
	cudaMemcpy( odata, g_odata, odataSize*reps, cudaMemcpyDeviceToHost);

	double timeTaken = wallTime;
	printf("Time taken: %.6fs\n",timeTaken);

	for(int i = 0; i < threads * blocks * reps; i++)
		timer[i] = (double)odata[i];
	
	mxFree(odata);
	CUDA_SAFE_CALL(cudaFree(g_odata));
	CUDA_SAFE_CALL(cudaFree(g_idata));

}

