#include <stdio.h>
#include <string.h>
#include "cuda.h"
#include "mex.h"
#include "../common/CudaHelper.h"


#define BLOCK_SIZE 384
#define DATA_PER_BLOCK (BLOCK_SIZE*4)
#define MAX_TAPS 50
#define MAX_BLOCKS 0xFFFF

template<int maxTaps> inline 
void executeKernel(dim3 dimGrid,dim3 dimBlock, unsigned int sharedMemSize, 
				   float *out_data_gpu, float *in_data_gpu, unsigned int data_len,
				   int taps)
{
	if(taps == maxTaps)
		convolution_kernel<maxTaps><<<dimGrid,dimBlock,sharedMemSize>>>( out_data_gpu, in_data_gpu, data_len);

	executeKernel<maxTaps-1>(dimGrid, dimBlock, sharedMemSize, 
				   out_data_gpu, in_data_gpu, data_len,taps);
}

template<> inline 
void executeKernel<1>(dim3 dimGrid, dim3 dimBlock, unsigned int sharedMemSize, 
				   float *out_data_gpu, float *in_data_gpu, unsigned int data_len,
				   int taps)
{
	return;
}

__constant__ float c_coeffs[MAX_TAPS];

template<int taps> __global__ void convolution_kernel( float *out_data, float *in_data, 
								  unsigned int data_len)
{
	extern __shared__ float shared[];

	const unsigned int ti = threadIdx.x;
	const unsigned int i = blockIdx.x*BLOCK_SIZE*4 + ti;
	float sum;

	// load data into shared memory
	shared[ti] = in_data[i];
	shared[BLOCK_SIZE + ti] = in_data[BLOCK_SIZE + i];
	shared[BLOCK_SIZE*2 + ti] = in_data[BLOCK_SIZE*2 + i];
	shared[BLOCK_SIZE*3 + ti] = in_data[BLOCK_SIZE*3 + i];

	if(ti < taps-1)
		shared[BLOCK_SIZE*4 + ti] = in_data[BLOCK_SIZE*4 + i];

	__syncthreads();

	sum = 0;
#pragma unroll
	for( int n = 0; n<taps; n++)
		sum += c_coeffs[n]*shared[ti+taps-1-n];
	out_data[i] = sum;

	sum = 0;
#pragma unroll
	for( int n = 0; n<taps; n++)
		sum += c_coeffs[n]*shared[BLOCK_SIZE + ti+taps-1-n];
	out_data[BLOCK_SIZE + i] = sum;

	sum = 0;
#pragma unroll
	for( int n = 0; n<taps; n++)
		sum += c_coeffs[n]*shared[BLOCK_SIZE*2 + ti+taps-1-n];
	out_data[BLOCK_SIZE*2 + i] = sum;

	sum = 0;
#pragma unroll
	for( int n = 0; n<taps; n++)
		sum += c_coeffs[n]*shared[BLOCK_SIZE*3 + ti+taps-1-n];
	out_data[BLOCK_SIZE*3 + i] = sum;
}

double convolution(float *in_data, float *out_data, float *coeffs, 
				 unsigned int data_len, unsigned int filter_len, unsigned int padding)
{

	// allocate memory
	float *out_data_gpu, *in_data_gpu;
	cudaMalloc( (void **) &out_data_gpu,sizeof(float)*(padding + data_len));
	cudaMalloc( (void **) &in_data_gpu,sizeof(float)*(padding + data_len + filter_len - 1));

	// copy to device
	cudaMemcpyToSymbol( c_coeffs, coeffs, sizeof(float)*filter_len);
	cudaMemcpy( in_data_gpu, in_data, sizeof(float)*(padding + data_len + filter_len - 1), cudaMemcpyHostToDevice);

	int sharedMemSize = sizeof(float)*(DATA_PER_BLOCK+filter_len-1);

	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( (data_len+padding)/DATA_PER_BLOCK );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);


	TIC();

	int reps = 100;
	for(int n=0;n<reps;n++){
		executeKernel<MAX_TAPS>(dimGrid, dimBlock, sharedMemSize, 
				   out_data_gpu, in_data_gpu, data_len, filter_len);
	}

	float time_taken;
	TOC(time_taken);

	printf("avg. time: %.2f(ms)\n",1000*time_taken/reps);

	double datatransferred = (double)sizeof(float)*reps*(2*(padding + data_len) + dimGrid.x*(filter_len - 1));
	
	printf("Transferrate: %.2fGiB/s\n",datatransferred/(time_taken*1024*1024*1024));

	double samples = (double)reps*data_len;
	printf("Filterrate: %.2fGS/s\n",samples/(time_taken*1024*1024*1024));


	// copy result to host memory
	cudaMemcpy( out_data, out_data_gpu, sizeof(float)*data_len, cudaMemcpyDeviceToHost);
	

	// clean up

	cudaFree(out_data_gpu);
	cudaFree(in_data_gpu);

	return time_taken/reps;
}


int length(const mxArray *vector){
	if( mxGetM(vector) == 1 ){
		return(mxGetN(vector));
	} else if( mxGetN(vector) == 1 ) {
		return(mxGetM(vector));
	} else {
		mexErrMsgTxt("Not a vector");
		return 0;
	}
}

/* Gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// Get parameters
	//
	// function [out_data] = convolution( coeffs, in_data )

	const mxArray *mx_coeffs =	prhs[0], 
		          *mx_in_data =	prhs[1];

	float *in_dataf, *out_dataf, *coeffsf;
	double *in_data, *out_data, *coeffs;
	unsigned int data_len, filter_len;


	// check validity of parameters
	if (nrhs != 2)
		mexErrMsgTxt("Wrong input parameters. convolution( coeffs, in_data )");
	
	data_len = length(mx_in_data);
	filter_len = length(mx_coeffs);

	if (filter_len > MAX_TAPS)
		mexErrMsgTxt("Filter too long.");

	if (data_len > MAX_BLOCKS*DATA_PER_BLOCK)
		mexErrMsgTxt("Data too long.");


	// get pointers
	in_data = mxGetPr(mx_in_data);
	coeffs = mxGetPr(mx_coeffs);

	// calculate padding
	int padding = ((data_len + DATA_PER_BLOCK - 1)/DATA_PER_BLOCK)*DATA_PER_BLOCK - data_len;
	printf("filter length = %d   data length = %d   padding = %d\n",filter_len, data_len, padding);

	// allocate memory
	in_dataf = (float*)mxMalloc(sizeof(float)*(padding + data_len + filter_len - 1));
	out_dataf = (float*)mxMalloc(sizeof(float)*(padding + data_len));
	coeffsf = (float*)mxMalloc(sizeof(float)*filter_len);
	
	// pad with zeros at beginning
	for( int i = 0; i < filter_len-1; i++)
		in_dataf[i] = 0.0f;

	// convert in_data. double -> float
	for( int i = 0; i < data_len; i++)
		in_dataf[i + filter_len - 1] = (float)in_data[i];
	
	// pad with zeros at end
	for( int i = 0; i < padding; i++)
		in_dataf[i + filter_len - 1 + data_len] = 0.0f;

	// convert coeffs. double -> float
	for( int i = 0; i < filter_len; i++)
		coeffsf[i] = (float)coeffs[i];

	

	// do it
	double ret_time = convolution(in_dataf, out_dataf, coeffsf, data_len, filter_len, padding);

	// create result matrix
	plhs[0] = mxCreateDoubleMatrix(data_len, 1, mxREAL);
	out_data = mxGetPr(plhs[0]);

	// convert out_data. float -> double
	for (int i = 0; i < data_len; i++)
			out_data[i] = (double)out_dataf[i];

	if(nlhs == 2){
		plhs[1] = mxCreateDoubleMatrix(1, 1, mxREAL);
		double *ret_time_p = mxGetPr(plhs[1]);
		ret_time_p[0] = ret_time;
	}

	// free memory
	mxFree(in_dataf);
	mxFree(out_dataf);
	mxFree(coeffsf);

}

