function g = delay_filter(D,N,K,fs,fc,rel_bandwidth)

% Comments refer to derivative filter that this delay filter
% is based on.
% Generates filter, g, of length K to differentiate signals
% of length N with centre frequency fc and sampling frquency fs.
% Relative bandwidth is assumed to be 50%
%
% Notice that for K=2, the output is not generally [-0.5 0 0.5], but
% rather some slightly different values since [-0.5 0 0.5] tends
% to underestimate the derivative for sine curves.
%

% OKVS 2005-12-15


if round(N/2)==N/2,
    % we should use odd length (nicer plots, otherwise unimportant)
    N = N+1;
end

silent = 1;
if fc==0,
   bandwidth = rel_bandwidth;
else
   if nargin <= 5,
       rel_bandwidth = 0.5;       
   end
   bandwidth = rel_bandwidth*fc;
end

n = (0:N-1)';
nc = fc/fs;
variance = bandwidth^2/(8*log(2));
x = round(N*(1/2-fc/fs))/N;

w = leftshift(gaussian(N,1,nc+x,variance),round(N*x));

w = (w + leftshift(flipud(w(:)),-1)')';
w = w/max(abs(w(:)));
W = sparse(diag(w));

hh = exp(-i*2*pi*D*[(0:ceil(N/2)-1)';(-floor(N/2):-1)']/N);

f = (-floor(N/2):floor((N-1)/2))/N*fs;
if ~silent,
   fp(f,fftshift(w)),title('Weights')
   figure,subplot(211),plot(f,fftshift(abs(hh)),'k-'),grid on;
   subplot(212),plot(f,fftshift(angle(hh)),'k-'),grid on;
end
for j = 1:length(K),
   
   k = (-floor(K(j)/2):floor((K(j)-1)/2))';
   F = exp(-i*2*pi/N*n*k');
   A = F'*W*F;
   
   b = F'*(w.*hh);
   g{j} = real(A\b);

   if ~silent,
       col=['rgbmcy'];
       subplot(211),hold on,title('Magnitude response');
       plot(f,fftshift(abs(F*g{j})),col(1+mod(j-1,length(col))));
       subplot(212),hold on,title('Phase response');
       fph(f,fftshift(angle(F*g{j})),[col(1+mod(j-1,length(col))),'.-']);
   end
end

if length(K)==1,% convert cell to vector
   g = g{:};
   g = g./sum(g);
end