#include <stdio.h>
#include <math.h>
#include "cuda.h"
#include "../common/CudaHelper.h"
#include "mex.h"

#define BLOCK_SIZE 256
#define MAX_CHANNELS 256
#define MAX_BEAMS 200
#define MAX_MLA 63
#define DOWNSAMPLE 4

texture<float,2,cudaReadModeElementType> tex;
texture<float,2,cudaReadModeNormalizedFloat> texApo;


__constant__ float C1_gpu[MAX_CHANNELS];
__constant__ float C3_gpu[MAX_CHANNELS];
__constant__ float C4_gpu[MAX_BEAMS][MAX_MLA];

__global__ void delay_and_sum_kernel( float *out_data, int channels, int samples, int no_mla )
{
	int idx = blockIdx.x*BLOCK_SIZE+threadIdx.x;
	if(idx < samples){
		float i = DOWNSAMPLE*idx;
		float apoX = 0.5f + idx;
		int beam = blockIdx.y;
		idx += beam*no_mla*samples;
		for(int mla = 0; mla < no_mla; mla++){
			float out = 0.0f;
			float y = 0.5f + channels*beam;
			float apoY = 0.5f;
			float y_max = channels*(beam+1);
			int n = 0;
			do{
				out += tex2D(texApo, apoX,apoY) *
					   tex2D(tex, 0.5f + i/2 + sqrt(i*i/4 + C1_gpu[n] - i*C3_gpu[n]*C4_gpu[beam][mla]),y);
				y += 1.0f;
				apoY += 1.0f;
				n += 1;
			} while( y < y_max );
			out_data[idx] = out;	
			idx += samples;
		}
	}
}


void checkLaunch(){
	cudaThreadSynchronize();                      
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) {                                          
        fprintf(stderr, "failed : %s.\n",
               cudaGetErrorString( err) );
    }
}

void delay_and_sum(float *in_data, float *out_data, short *apo, double *a_c, double f_s, double c_0, 
				   double *theta, int samples_in, int samples_out, int channels, 
				   int beams, int no_mla)
{
	double f_s_up = f_s*DOWNSAMPLE;

	// allocate memory
	float *out_data_gpu;
	CUDA_SAFE_CALL(cudaMalloc( (void **) &out_data_gpu,sizeof(float)*samples_out*beams*no_mla));


	float C1[MAX_CHANNELS] = {0};
	float C3[MAX_CHANNELS] = {0};
	float C4[MAX_BEAMS][MAX_MLA] = {0};

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);

	// allocate device memory in_data for texture
	cudaArray* in_data_gpu;
	CUDA_SAFE_CALL(cudaMallocArray( &in_data_gpu, &channelDesc, samples_in, channels*beams )); 

	// copy host data to device array
	CUDA_SAFE_CALL(cudaMemcpyToArray( in_data_gpu, 0, 0, in_data, sizeof(float)*samples_in*channels*beams, cudaMemcpyHostToDevice));

	// set texture parameters
	tex.addressMode[0] = cudaAddressModeClamp;
	tex.addressMode[1] = cudaAddressModeClamp;
	tex.filterMode = cudaFilterModeLinear;
	tex.normalized = false;

	// Bind the array to the texture
	CUDA_SAFE_CALL(cudaBindTextureToArray(tex, in_data_gpu, channelDesc));


	// APODIZE texture
	cudaChannelFormatDesc channelDescApo = cudaCreateChannelDesc(16, 0, 0, 0, cudaChannelFormatKindSigned);
	cudaArray* apo_gpu;
	// allocate device memory for apodize texture
	cudaMallocArray( &apo_gpu, &channelDescApo, samples_out, channels ); 

	// set texture parameters
	texApo.addressMode[0] = cudaAddressModeClamp;
	texApo.addressMode[1] = cudaAddressModeClamp;
	texApo.filterMode = cudaFilterModePoint;
	texApo.normalized = false;

	// copy host data to device array
	CUDA_SAFE_CALL(cudaMemcpyToArray( apo_gpu, 0, 0, apo, sizeof(short)*samples_out*channels, cudaMemcpyHostToDevice));


	// Bind the array to the texture
	cudaBindTextureToArray(texApo, apo_gpu, channelDescApo);



	// calculate C1, C3 and C4
	double lambda = f_s_up/c_0;
	for (int j = 0; j < channels; j++) {
		double temp = a_c[j]*lambda;
		C1[j] = (float)( temp*temp );
		C3[j] = (float)temp;
	}	
	for (int b = 0; b < beams; b++) {
		for (int mla = 0; mla < no_mla; mla++) {
			C4[b][mla] = (float)sin(theta[b*no_mla + mla]);
		}
	}

	// copy C1, C3 and C4 to device
	cudaMemcpyToSymbol( C1_gpu, C1, sizeof(float)*MAX_CHANNELS);
	cudaMemcpyToSymbol( C3_gpu, C3, sizeof(float)*MAX_CHANNELS);
	cudaMemcpyToSymbol( C4_gpu, C4, sizeof(float)*MAX_MLA*MAX_BEAMS);


	// run kernel
	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( (samples_out+BLOCK_SIZE-1)/BLOCK_SIZE,beams,1 );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);

	TIC();

	int reps = 100;
	for(int n=0;n<reps;n++)
		delay_and_sum_kernel<<<dimGrid,dimBlock>>>( out_data_gpu, channels, samples_out, no_mla );

	float wallTime;
	TOC(wallTime);

	checkLaunch();

	printf("Repetitions = %d   Walltime = %fs\n",reps,wallTime);

	float dataTransfered = (float)reps*beams*(channels*sizeof(float)*samples_in + sizeof(float)*samples_out);
	printf("Speed: %.2fx\n",((double)beams*samples_out/f_s)/(wallTime/reps));
	printf("Speed per MLA: %.2fx\n",((double)no_mla*beams*samples_out/f_s)/(wallTime/reps));
	printf("Datatransfer: at least %.2fGB/s\n",dataTransfered/(wallTime*1024*1024*1024));
	double threadsCompleted = (double)reps * dimBlock.x * dimBlock.y * dimBlock.z * dimGrid.x * dimGrid.y * dimGrid.z;
	double threadsInparallel = 128;
	double SecondsPerThread = threadsInparallel*wallTime/threadsCompleted;
	double ticksPerSecond = 1.35e9;
	printf("Clock cycles to complete a thread: approx. %.1fticks\n",SecondsPerThread*ticksPerSecond);



	// copy result to host memory
	CUDA_SAFE_CALL(cudaMemcpy( out_data, out_data_gpu, sizeof(float)*samples_out*beams*no_mla, cudaMemcpyDeviceToHost));
	
	// clean up
	CUDA_SAFE_CALL(cudaUnbindTexture(tex));
	CUDA_SAFE_CALL(cudaUnbindTexture(texApo));
	CUDA_SAFE_CALL(cudaFreeArray(in_data_gpu));
	CUDA_SAFE_CALL(cudaFreeArray(apo_gpu));


	CUDA_SAFE_CALL(cudaFree(out_data_gpu));
}


int length(const mxArray *vector){
	if( mxGetM(vector) == 1 ){
		return(mxGetN(vector));
	} else if( mxGetN(vector) == 1 ) {
		return(mxGetM(vector));
	} else {
		mexErrMsgTxt("Not a vector");
		return 0;
	}
}

/* Gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// Get parameters
	//
	// function [out_data] = quarterTEXF_MLA_APO( in_data, a_c, f_s, c_0, theta, apo )

	const mxArray *mx_in_data =	prhs[0], 
		          *mx_a_c =		prhs[1], 
				  *mx_f_s =		prhs[2], 
				  *mx_c_0 =		prhs[3],
				  *mx_theta =	prhs[4],
				  *mx_apo =		prhs[5];


	// check validity of parameters
	if (nrhs != 6)
		mexErrMsgTxt("Wrong input parameters. quarterTEXF_MLA_APO( in_data, a_c, f_s, c_0, theta, apo )\n(f_s is output samplerate)");

	int channels = length(mx_a_c);
	int samples_in = mxGetM(mx_in_data);
	int samples_out = samples_in/DOWNSAMPLE;
	int beams = mxGetN(mx_theta);
	int no_mla = mxGetM(mx_theta);

	
	if( mxGetN(mx_in_data) != channels*beams )
		mexErrMsgTxt("Number of columns of in_data must equal length of a_c times length of theta");

	if( no_mla > MAX_MLA )
		mexErrMsgTxt("Number of MLAs must be no more than 63");

	if( beams > MAX_BEAMS )
		mexErrMsgTxt("Number of beams must be no more than 200");

	if( channels > MAX_CHANNELS )
		mexErrMsgTxt("Number of channels must be no more than 256");

	if( mxGetM(mx_apo) != samples_out )
		mexErrMsgTxt("Number of rows in apo must be 1/4 as many as in in_data");

	if( mxGetN(mx_apo) != channels )
		mexErrMsgTxt("Number of columns in apo must be same as number of channels");

	printf("%d samples(out) x %d channels x %d mla x %d beams\n",samples_out,channels, no_mla, beams);

	// get scalars
	double f_s = mxGetScalar(mx_f_s);
	double c_0 = mxGetScalar(mx_c_0);

	// get pointers
	double *in_data = mxGetPr(mx_in_data);
	double *a_c = mxGetPr(mx_a_c);
	double *theta = mxGetPr(mx_theta);
	double *apo = mxGetPr(mx_apo);




	// allocate memory
	float *in_dataf = (float*)mxMalloc(sizeof(float)*samples_in*channels*beams);
	float *out_dataf = (float*)mxMalloc(sizeof(float)*samples_out*beams*no_mla);
	short *apos = (short*)mxMalloc(sizeof(short)*samples_out*channels);

	// convert in_data. double -> float
	for( int i = 0; i < samples_in*channels*beams; i++)
		in_dataf[i] = (float)in_data[i];

	// convert apo. double -> short
	for( int i = 0; i < samples_out*channels; i++)
		apos[i] = (short)(apo[i]*0x7fff);

	// do it
	delay_and_sum(in_dataf, out_dataf, apos, a_c, f_s, c_0, theta, samples_in, samples_out , channels, beams, no_mla);

	// create result matrix
	plhs[0] = mxCreateDoubleMatrix(samples_out, beams*no_mla, mxREAL);
	double *out_data = mxGetPr(plhs[0]);

	// convert out_data. float -> double


	for (int i = 0; i < samples_out*beams*no_mla; i++)
		out_data[i] = (double)out_dataf[i];

	// free memory
	mxFree(apos);
	mxFree(out_dataf);
	mxFree(in_dataf);

}