#include <stdio.h>
#include <math.h>
#include "cuda.h"
#include "../common/CudaHelper.h"
#define mxFree free
#define mxMalloc malloc

#define BLOCK_SIZE 384
#define NO_CHANNELS 128
#define SAMPLES (6144)
#define BEAMS 40
#define REPS 1000

texture<float,2,cudaReadModeNormalizedFloat> tex;

__constant__ float C_gpu[BEAMS][2][NO_CHANNELS];

__global__ void delay_and_sum_kernel( float *out_data )
{
	int idx = blockIdx.x*BLOCK_SIZE+threadIdx.x;
	float i = idx;
	float out = 0.0f;
	int beam = blockIdx.y;
	float y = 0.5f + NO_CHANNELS*beam;
	float y_max = NO_CHANNELS*(beam+1);
	int n = 0;
	do{
		out += tex2D(tex, 0.5f + i/2 + sqrt(i*i/4 + C_gpu[beam][0][n] - i*C_gpu[beam][1][n]),y);
		y += 1.0f;
		n += 1;
	} while( y < y_max );
	idx += beam*SAMPLES;
	out_data[idx] = out;	
}


void delay_and_sum(double *a_c, double f_s, double c_0, double *theta, 
				   unsigned int N, unsigned int no_channels)
{

	// allocate memory
	float *out_data_gpu;
	cudaMalloc( (void **) &out_data_gpu,sizeof(float)*N*BEAMS);


	float C[BEAMS][2][NO_CHANNELS];

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(16, 0, 0, 0, cudaChannelFormatKindSigned);

	cudaArray* in_data_gpu;

	// allocate device memory in_data for texture
	cudaMallocArray( &in_data_gpu, &channelDesc, N, no_channels*BEAMS ); 

	// set texture parameters
	tex.addressMode[0] = cudaAddressModeClamp;
	tex.addressMode[1] = cudaAddressModeClamp;
	tex.filterMode = cudaFilterModeLinear;
	tex.normalized = false;

	// Bind the array to the texture
	cudaBindTextureToArray(tex, in_data_gpu, channelDesc);



	// calculate C1 and C2
	double lambda = f_s/c_0;
	for (int b = 0; b < BEAMS; b++) {
		for (int j = 0; j < no_channels; j++) {
			double temp = a_c[j]*lambda;
			C[b][0][j] = (float)( temp*temp );
			C[b][1][j] = (float)( temp*sin(theta[b]) );
		}	
	}

	// copy C1 and C2 to device
	cudaMemcpyToSymbol( C_gpu, C, sizeof(float)*NO_CHANNELS*2);



	// run kernel
	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( N/BLOCK_SIZE,BEAMS,1 );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);

	TIC();

	int reps = REPS;
	for(int n=0;n<reps;n++)
		delay_and_sum_kernel<<<dimGrid,dimBlock>>>( out_data_gpu );

	float wallTime;
	TOC(wallTime);

	float dataTransfered = (float)reps*BEAMS*N*(no_channels*sizeof(short) + sizeof(float));
	printf("time avg per beam: %.2f(ms)\n",1000*wallTime/(reps*BEAMS));
	printf("time available per beam: %.2f(ms)\n",1000*N/f_s);
	printf("Speed: %.2fx\n",(BEAMS*N/f_s)/(wallTime/reps));
	printf("Datatransfer: approx. %.2fGB/s\n",dataTransfered/(wallTime*1024*1024*1024));
	double threadsCompleted = (double)reps * dimBlock.x * dimBlock.y * dimBlock.z * dimGrid.x * dimGrid.y * dimGrid.z;
	double threadsInparallel = 128;
	double SecondsPerThread = threadsInparallel*wallTime/threadsCompleted;
	double ticksPerSecond = 1.35e9;
	printf("Clock ticks to complete a thread: approx. %.1fticks\n",SecondsPerThread*ticksPerSecond);
	double samplesProcessed = (double)no_channels*threadsCompleted;
	printf("Filterrate: %.2f*10^6 samples in per second\n",samplesProcessed/(wallTime*1000*1000));


	printf("Channels = %d  samples = %d  beams = %d  Repetitions = %d\n",NO_CHANNELS, N, BEAMS, reps);
	printf("Walltime = %fs\n",wallTime);
	
	// clean up
	cudaUnbindTexture(tex);
	cudaFreeArray(in_data_gpu);

	cudaFree(out_data_gpu);
}

int main()
{
	printf("TEXF\n");
	double *a_c, *theta;
	double f_s,c_0;
	unsigned int no_channels, N;

	N = SAMPLES;
	no_channels = NO_CHANNELS;

	// get scalars
	f_s = 40000000;
	c_0 = 1504;

	// allocate memory
	a_c = (double*)malloc(sizeof(double)*no_channels);
	theta = (double*)malloc(sizeof(double)*BEAMS);

	// generate a_c. 
	for( int i = 0; i < no_channels; i++)
		a_c[i] = (i-no_channels/2)*0.0002;

	// generate theta. 
	for( int i = 0; i < BEAMS; i++)
		theta[i] = (i-BEAMS/2.0)*(1.047/BEAMS); // pi/3 sweep

	// do it
	delay_and_sum(a_c, f_s, c_0, theta, N, no_channels);


	// free memory
	free(theta);
	free(a_c);
}

