#include <stdio.h>
#include <math.h>
#include "cuda.h"
#include "../common/CudaHelper.h"
#define mxFree free
#define mxMalloc malloc

#define BLOCK_SIZE 64
#define NO_CHANNELS 128
#define SAMPLES 6144
#define BEAMS 40

texture<float,2,cudaReadModeNormalizedFloat> tex;
texture<float,2,cudaReadModeNormalizedFloat> texApo;

__constant__ float C1_gpu[NO_CHANNELS];
__constant__ float C3_gpu[NO_CHANNELS];
__constant__ float C4_gpu[BEAMS];

__global__ void delay_and_sum_kernel( float *out_data )
{
	int idx = blockIdx.x*BLOCK_SIZE+threadIdx.x;
	float i = idx;
	float out = 0.0f;
	int beam = blockIdx.y;
	float y = 0.5f + NO_CHANNELS*beam;
	float apoY = 0.5f;
	float y_max = NO_CHANNELS*(beam+1);
	int n = 0;
	do{
		out += tex2D(texApo, 0.5f + i,apoY) *
			tex2D(tex, 0.5f + i/2 + sqrt(i*i/4 + C1_gpu[n] - i*C3_gpu[n]*C4_gpu[beam]),y);
		y += 1.0f;
		apoY += 1.0f;
		n += 1;
	} while( y < y_max );
	idx += beam*SAMPLES;
	out_data[idx] = out;	
}


void delay_and_sum(double *a_c, double f_s, double c_0, double *theta, 
				   unsigned int N, unsigned int no_channels)
{

	// allocate memory
	float *out_data_gpu;
	cudaMalloc( (void **) &out_data_gpu,sizeof(float)*N*BEAMS);


	float C1[NO_CHANNELS] = {0};
	float C3[NO_CHANNELS] = {0};
	float C4[BEAMS] = {0};

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(16, 0, 0, 0, cudaChannelFormatKindSigned);

	cudaArray* in_data_gpu;

	// allocate device memory in_data for texture
	cudaMallocArray( &in_data_gpu, &channelDesc, N, no_channels*BEAMS ); 

	// set texture parameters
	tex.addressMode[0] = cudaAddressModeClamp;
	tex.addressMode[1] = cudaAddressModeClamp;
	tex.filterMode = cudaFilterModeLinear;
	tex.normalized = false;

	// Bind the array to the texture
	cudaBindTextureToArray(tex, in_data_gpu, channelDesc);


	// APODIZE texture
	cudaChannelFormatDesc channelDescApo = cudaCreateChannelDesc(16, 0, 0, 0, cudaChannelFormatKindSigned);
	cudaArray* apo_gpu;
	// allocate device for apodize texture
	cudaMallocArray( &apo_gpu, &channelDescApo, N, no_channels ); 

	// set texture parameters
	texApo.addressMode[0] = cudaAddressModeClamp;
	texApo.addressMode[1] = cudaAddressModeClamp;
	texApo.filterMode = cudaFilterModePoint;
	texApo.normalized = false;

	// Bind the array to the texture
	cudaBindTextureToArray(texApo, apo_gpu, channelDescApo);




	// calculate C1, C3 and C4
	double lambda = f_s/c_0;
	for (int j = 0; j < no_channels; j++) {
		double temp = a_c[j]*lambda;
		C1[j] = (float)( temp*temp );
		C3[j] = (float)temp;
	}	
	for (int b = 0; b < BEAMS; b++) {
		C4[b] = (float)sin(theta[b]);
	}

	// copy C1, C3 and C4 to device
	cudaMemcpyToSymbol( C1_gpu, C1, sizeof(float)*NO_CHANNELS);
	cudaMemcpyToSymbol( C3_gpu, C3, sizeof(float)*NO_CHANNELS);
	cudaMemcpyToSymbol( C4_gpu, C4, sizeof(float)*BEAMS);



	// run kernel
	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( N/BLOCK_SIZE,BEAMS,1 );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);

	TIC();

	int reps = 1000;
	for(int n=0;n<reps;n++)
		delay_and_sum_kernel<<<dimGrid,dimBlock>>>( out_data_gpu );

	float wallTime;
	TOC(wallTime);

	float dataTransfered = (float)reps*BEAMS*N*(2*no_channels*sizeof(short) + sizeof(float));
	printf("time avg per beam: %.2f(ms)\n",1000*wallTime/(reps*BEAMS));
	printf("time available per beam: %.2f(ms)\n",1000*N/f_s);
	printf("Speed: %.2fx\n",(BEAMS*N/f_s)/(wallTime/reps));
	printf("Datatransfer: approx. %.2fGB/s\n",dataTransfered/(wallTime*1024*1024*1024));
	double threadsCompleted = (double)reps * dimBlock.x * dimBlock.y * dimBlock.z * dimGrid.x * dimGrid.y * dimGrid.z;
	double threadsInparallel = 128;
	double SecondsPerThread = threadsInparallel*wallTime/threadsCompleted;
	double ticksPerSecond = 1.35e9;
	printf("Clock ticks to complete a thread: approx. %.1fticks\n",SecondsPerThread*ticksPerSecond);
	double samplesProcessed = (double)no_channels*threadsCompleted;
	printf("Filterrate: %.2f*10^6 samples in per second\n",samplesProcessed/(wallTime*1000*1000));

	printf("Channels = %d  samples = %d  beams = %d  Repetitions = %d\n",NO_CHANNELS, N, BEAMS, reps);
	printf("Walltime = %fs\n",wallTime);
	
	// clean up
	cudaUnbindTexture(tex);
	cudaUnbindTexture(texApo);
	cudaFreeArray(in_data_gpu);
	cudaFreeArray(apo_gpu);

	cudaFree(out_data_gpu);
}

int main()
{
	printf("TEXF APO\n");
	double *a_c, *theta;
	double f_s,c_0;
	unsigned int no_channels, N;

	N = SAMPLES;
	no_channels = NO_CHANNELS;

	// get scalars
	f_s = 40000000;
	c_0 = 1504;

	// allocate memory
	a_c = (double*)malloc(sizeof(double)*no_channels);
	theta = (double*)malloc(sizeof(double)*BEAMS);

	// generate a_c. 
	for( int i = 0; i < no_channels; i++)
		a_c[i] = (i-no_channels/2)*0.0002;

	// generate theta. 
	for( int i = 0; i < BEAMS; i++)
		theta[i] = (i-BEAMS/2.0)*(1.047/BEAMS); // pi/3 sweep

	// do it
	delay_and_sum(a_c, f_s, c_0, theta, N, no_channels);


	// free memory
	free(theta);
	free(a_c);
}

