#include <stdio.h>
#include <string.h>
#include "cuda.h"
#define mxMalloc malloc
#include "../common/CudaHelper.h"

#define COALESCED_PITCH 1

#define BLOCK_SIZE 256
// DATA_PER_BLOCK: each thread handles two samples
#define DATA_PER_THREAD 2
#define DATA_PER_BLOCK (DATA_PER_THREAD*BLOCK_SIZE)

// TAPS must be odd number
#define TAPS 11

// SAMPLES should be a multiple of DATA_PER_BLOCK
#define SAMPLES (2048*16)
#define CHANNELS 128

__constant__ float c_coeffs[3][TAPS];

texture<float2, 1, cudaReadModeNormalizedFloat> tex_short2;

__global__ void upsample4x_kernel( float4 *out_data, short channel_pitch_in, short channel_pitch_out )
{
	extern __shared__ float shared[];

	const unsigned int ti = threadIdx.x;
	const unsigned int i_out = blockIdx.y*channel_pitch_out + blockIdx.x*BLOCK_SIZE*2 + ti;
	const unsigned int i_in = blockIdx.y*channel_pitch_in + blockIdx.x*BLOCK_SIZE + ti;
	float4 out;

	// load data into shared memory
	float2 in = tex1Dfetch(tex_short2, i_in);
	shared[ti*2]   = in.x; 
	shared[ti*2+1] = in.y;

	if(ti < TAPS-1){
		in = tex1Dfetch(tex_short2, BLOCK_SIZE + i_in);
		shared[BLOCK_SIZE*2 + ti*2]   = in.x; 
		shared[BLOCK_SIZE*2 + ti*2+1] = in.y;
	}

	__syncthreads();


	out.x = shared[ti + TAPS/2];	// D = 0
	out.y = 0.0f; out.z = 0.0f; out.w = 0.0f;
#pragma unroll
	for( int n = 0; n<TAPS; n++)
		out.y += c_coeffs[0][n]*shared[ti+TAPS-1-n]; // D = -0.25
#pragma unroll
	for( int n = 0; n<TAPS; n++)
		out.z += c_coeffs[1][n]*shared[ti+TAPS-1-n]; // D = -0.5
#pragma unroll
	for( int n = 0; n<TAPS; n++)
		out.w += c_coeffs[2][n]*shared[ti+TAPS-1-n]; // D = -0.75
	out_data[i_out] = out;

	out.x = shared[BLOCK_SIZE + ti + TAPS/2];	// D = 0
	out.y = 0.0f; out.z = 0.0f; out.w = 0.0f;
#pragma unroll
	for( int n = 0; n<TAPS; n++)
		out.y += c_coeffs[0][n]*shared[BLOCK_SIZE + ti+TAPS-1-n]; // D = -0.25
#pragma unroll
	for( int n = 0; n<TAPS; n++)
		out.z += c_coeffs[1][n]*shared[BLOCK_SIZE + ti+TAPS-1-n]; // D = -0.5
#pragma unroll
	for( int n = 0; n<TAPS; n++)
		out.w += c_coeffs[2][n]*shared[BLOCK_SIZE + ti+TAPS-1-n]; // D = -0.75
	
	out_data[BLOCK_SIZE + i_out] = out;
}


void upsample4x(short *in_data, float *out_data, float *coeffs, 
				 unsigned int data_len, unsigned int filter_len, unsigned int padding,
				 unsigned int padding2)
{
	// allocate memory
	float4 *out_data_gpu;
	short *in_data_gpu;
	int channel_pitch_in = padding + padding2 + data_len + filter_len - 1;
	int in_data_len = sizeof(short)*channel_pitch_in*CHANNELS;
	int channel_pitch_out = padding + data_len;
	int out_data_len = sizeof(float4)*channel_pitch_out*CHANNELS;
	cudaMalloc( (void **) &in_data_gpu, in_data_len);
	cudaMalloc( (void **) &out_data_gpu, out_data_len);

	// copy to device
	//cudaMemcpyToSymbol( c_coeffs, coeffs, sizeof(float)*filter_len*3);
	//cudaMemcpy( in_data_gpu, in_data, sizeof(short)*(padding + data_len + filter_len - 1), cudaMemcpyHostToDevice);

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(16, 16, 0, 0, cudaChannelFormatKindSigned);
	CUDA_SAFE_CALL( cudaBindTexture(0, tex_short2, in_data_gpu, channelDesc, in_data_len ));

	int sharedMemSize = sizeof(float)*(DATA_PER_BLOCK+DATA_PER_THREAD*(filter_len-1));

	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( (data_len+padding)/DATA_PER_BLOCK,CHANNELS,1 );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);

    TIC();

	int reps = 1000;
	for(int n=0;n<reps;n++){
		upsample4x_kernel<<<dimGrid,dimBlock,sharedMemSize>>>( out_data_gpu, channel_pitch_in, channel_pitch_out );
	}

	float time_taken;
	TOC(time_taken);

	printf("Repitions = %d\n",reps);
	printf("Walltime = %fs\n",time_taken);

	printf("avg. time: %.3f(ms)\n",1000*time_taken/reps);

	double datatransferred = (double)reps*( dimGrid.x*dimGrid.y*sizeof(short2)*dimBlock.x + 
		out_data_len );
	printf("Transferrate: (at least) %.2fGiB/s\n",datatransferred/(time_taken*1024*1024*1024));

	double samples = (double)CHANNELS*reps*data_len;
	printf("Filterrate: %.2f*10^6 samples in per second\n",samples/(time_taken*1000*1000));

	double FLOPS = (double)reps*data_len*filter_len*2*3*CHANNELS;
	printf("Floating point operation rate: %.2fGFLOPS\n",FLOPS/(time_taken*1000*1000*1000));

	double threadsCompleted =  dimBlock.x * dimBlock.y * dimBlock.z * dimGrid.x * dimGrid.y * dimGrid.z;
	double threadsInparallel = 128;
	double SecondsPerThread = threadsInparallel*(time_taken/reps)/threadsCompleted;
	double ticksPerSecond = 1.35e9;
	printf("Clock ticks to complete a thread: approx. %.1fticks\n",SecondsPerThread*ticksPerSecond);



	// copy result to host memory
	//cudaMemcpy( out_data, out_data_gpu, sizeof(float)*data_len, cudaMemcpyDeviceToHost);
	

	// clean up

	cudaFree(out_data_gpu);
	cudaFree(in_data_gpu);
}


/* Gateway function */
int main(int argc, char *args[])
{
	// Get parameters
	//
	// function [out_data] = convolution( coeffs, in_data )


	short *in_datas;
	float *out_dataf, *coeffsf;
//	double *in_data, *out_data, *coeffs;
	unsigned int data_len, filter_len;


	data_len = SAMPLES;
	filter_len = TAPS;


	// calculate padding
	int padding = ((data_len + DATA_PER_BLOCK - 1)/(DATA_PER_BLOCK))*DATA_PER_BLOCK - data_len;
	int read_len = (padding + data_len + filter_len - 1);
	int padding2 = ((read_len + COALESCED_PITCH - 1)/(COALESCED_PITCH))*COALESCED_PITCH - read_len;
	printf("filter length = %d   data length = %d   padding = %d   padding2 = %d\n",filter_len, data_len, padding, padding2);

	int channel_pitch = read_len+padding2;

	// allocate memory
	in_datas = (short*)mxMalloc(sizeof(short)*channel_pitch*CHANNELS);
	out_dataf = (float*)mxMalloc(sizeof(float)*4*(padding + data_len)*CHANNELS);
	coeffsf = (float*)mxMalloc(sizeof(float)*filter_len*3);
	
	for( int j = 0; j < CHANNELS; j++){
		// pad with zeros at beginning
		for( int i = 0; i < (filter_len/2); i++)
			in_datas[j*channel_pitch + i] = 0;

		// convert in_data. double -> float
		for( int i = 0; i < data_len; i++)
			in_datas[j*channel_pitch + i + filter_len/2] = (short)0x8000*(float)(rand()/RAND_MAX)*2 - 1;
		
		// pad with zeros at end
		for( int i = 0; i < padding+filter_len/2+padding2; i++)
			in_datas[j*channel_pitch + i + filter_len/2 + data_len] = 0;
	}

	// convert coeffs. double -> float
	for( int j = 0; j < 3; j++)
		for( int i = 0; i < filter_len; i++)
			coeffsf[j*filter_len + i] = (float)(rand()/RAND_MAX)*2 - 1;	

	// do it
	upsample4x(in_datas, out_dataf, coeffsf, data_len, filter_len, padding, padding2);

	// free memory
	free(coeffsf);
	free(out_dataf);
	free(in_datas);

}

