#include <stdio.h>
#include <math.h>
#include "cuda.h"
#include "../common/CudaHelper.h"
#define mxFree free
#define mxMalloc malloc

#define BLOCK_SIZE 384
#define NO_CHANNELS 128
#define SAMPLES 6144
#define BEAMS 40
#define NO_MLA 16
#define DOWNSAMPLE 4

texture<float,2,cudaReadModeElementType> tex;

__constant__ float C1_gpu[NO_CHANNELS];
__constant__ float C3_gpu[NO_CHANNELS];
__constant__ float C4_gpu[BEAMS][NO_MLA];

__global__ void delay_and_sum_kernel( float *out_data )
{
	int idx = blockIdx.x*BLOCK_SIZE+threadIdx.x;
	float i = DOWNSAMPLE*idx;
	int beam = blockIdx.y;
	idx += beam*NO_MLA*SAMPLES;
	for(int mla = 0; mla < NO_MLA; mla++)
	{
		float out = 0.0f;
		float y = 0.5f + NO_CHANNELS*beam;
		float y_max = NO_CHANNELS*(beam+1);
		int n = 0;
		do{
			out += tex2D(tex, 0.5f + i/2 + sqrt(i*i/4 + C1_gpu[n] - i*C3_gpu[n]*C4_gpu[beam][mla]),y);
			y += 1.0f;
			n += 1;
		} while( y < y_max );
		out_data[idx] = out;	
		idx += SAMPLES;
	}
}


void delay_and_sum(double *a_c, double f_s, double c_0, double *theta, 
				   unsigned int samples_out, unsigned int no_channels)
{
	int samples_in = samples_out*DOWNSAMPLE;
	double f_s_up = f_s*DOWNSAMPLE;


	// allocate memory
	float *out_data_gpu;
	cudaMalloc( (void **) &out_data_gpu,sizeof(float)*samples_out*BEAMS);


	float C1[NO_CHANNELS] = {0};
	float C3[NO_CHANNELS] = {0};
	float C4[BEAMS][NO_MLA] = {0};

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);

	cudaArray* in_data_gpu;

	// allocate device memory in_data for texture
	cudaMallocArray( &in_data_gpu, &channelDesc, samples_in, no_channels*BEAMS ); 

	// set texture parameters
	tex.addressMode[0] = cudaAddressModeClamp;
	tex.addressMode[1] = cudaAddressModeClamp;
	tex.filterMode = cudaFilterModeLinear;
	tex.normalized = false;

	// Bind the array to the texture
	cudaBindTextureToArray(tex, in_data_gpu, channelDesc);



	// calculate C1, C3 and C4
	double lambda = f_s_up/c_0;
	for (int j = 0; j < no_channels; j++) {
		double temp = a_c[j]*lambda;
		C1[j] = (float)( temp*temp );
		C3[j] = (float)temp;
	}	
	for (int b = 0; b < BEAMS; b++) {
		for (int mla = 0; mla < NO_MLA; mla++) {
			C4[b][mla] = (float)sin(theta[b*NO_MLA + mla]);
		}
	}

	// copy C1, C3 and C4 to device
	cudaMemcpyToSymbol( C1_gpu, C1, sizeof(float)*NO_CHANNELS);
	cudaMemcpyToSymbol( C3_gpu, C3, sizeof(float)*NO_CHANNELS);
	cudaMemcpyToSymbol( C4_gpu, C4, sizeof(float)*NO_MLA*BEAMS);



	// run kernel
	dim3 dimBlock( BLOCK_SIZE );
	dim3 dimGrid( samples_out/BLOCK_SIZE,BEAMS,1 );

	printf("Block(%d,%d,%d), Grid(%d,%d,%d)\n", dimBlock.x, dimBlock.y, dimBlock.z, dimGrid.x, dimGrid.y, dimGrid.z);


	TIC();

	int reps = 100;
	for(int n=0;n<reps;n++)
		delay_and_sum_kernel<<<dimGrid,dimBlock>>>( out_data_gpu );

	float wallTime;
	TOC(wallTime);

	printf("Walltime = %fs\n",wallTime);
	float dataTransfered = (float)reps*BEAMS*(samples_in*no_channels*sizeof(float) + samples_out*sizeof(float));
	printf("Speed: %.2fx\n",(BEAMS*samples_out/f_s)/(wallTime/reps));
	printf("Speed per MLA: %.2fx\n",(NO_MLA*BEAMS*samples_out/f_s)/(wallTime/reps));
	printf("Datatransfer: approx. %.2fGB/s\n",dataTransfered/(wallTime*1024*1024*1024));
	double threadsCompleted = (double)reps * dimBlock.x * dimBlock.y * dimBlock.z * dimGrid.x * dimGrid.y * dimGrid.z;
	double threadsInparallel = 128;
	double SecondsPerThread = threadsInparallel*wallTime/threadsCompleted;
	double ticksPerSecond = 1.35e9;
	printf("Clock ticks to complete a thread: approx. %.1fticks\n",SecondsPerThread*ticksPerSecond);
	double samplesProcessed = (double)no_channels*threadsCompleted;
	printf("Filterrate: %.2f*10^6 samples in per second\n",samplesProcessed/(wallTime*1000*1000));

	printf("Channels = %d  samples(out) = %d  beams = %d  Repetitions = %d  MLAs = %d\n",NO_CHANNELS, samples_out, BEAMS, reps, NO_MLA);

	// clean up
	cudaUnbindTexture(tex);
	cudaFreeArray(in_data_gpu);

	cudaFree(out_data_gpu);
}

int main()
{
	printf("1/4 TEXF MLA\n");

	double *a_c, *theta;
	double f_s,c_0;
	unsigned int no_channels, N;

	N = SAMPLES;
	no_channels = NO_CHANNELS;

	// get scalars
	f_s = 40000000;
	c_0 = 1504;

	int beams_total = NO_MLA*BEAMS;

	// allocate memory
	a_c = (double*)malloc(sizeof(double)*no_channels);
	theta = (double*)malloc(sizeof(double)*beams_total);

	// generate a_c. 
	for( int i = 0; i < no_channels; i++)
		a_c[i] = (i-no_channels/2)*0.0002;

	// generate theta. 
	for( int i = 0; i < beams_total; i++)
		theta[i] = (i-beams_total/2.0)*(1.047/beams_total); // pi/3 sweep

	// do it
	delay_and_sum(a_c, f_s, c_0, theta, N, no_channels);


	// free memory
	free(theta);
	free(a_c);
}

